@@ -63,6 +63,8 @@ typedef struct {
 	uint32_t handled_requests;
 	time_t start_time;
 	char restart_issued;
+	int active_requests;
+	int waiting_requests;
 } dashboard_data;
 
 typedef struct xsp_data {
@@ -90,18 +92,25 @@ typedef struct xsp_data {
 	char is_virtual; /* is the server virtual? */
 	char *start_attempts;
 	char *start_wait_time;
+	char *max_active_requests;
+	char *max_waiting_requests;
   
 	/* auto-restart stuff */
 	auto_restart_mode restart_mode;
 	uint32_t restart_requests;
 	uint32_t restart_time;
+
+	/* other settings */
+	unsigned char no_flush;
 	
+#ifndef APACHE13
 	apr_shm_t *dashboard_shm;
 	dashboard_data *dashboard;
 	apr_global_mutex_t *dashboard_mutex;
 	char dashboard_mutex_initialized_in_child;
 	char *dashboard_file;
 	char *dashboard_lock_file;
+#endif
 } xsp_data;
 
 typedef struct {
@@ -116,6 +125,7 @@ typedef struct {
 	char *client_block_buffer;
 } request_data;
 
+#ifndef APACHE13
 typedef struct {
 	char *name;
 	apr_lockmech_e sym;
@@ -133,11 +143,13 @@ static lock_mechanism lockMechanisms [] = {
 	{"DEFAULT", APR_LOCK_DEFAULT, 1},
 	{NULL, 0, 0}
 };
+#endif
 
 static int send_table (apr_pool_t *pool, apr_table_t *table, apr_socket_t *sock);
 static void start_xsp (module_cfg *config, int is_restart, char *alias);
 static apr_status_t terminate_xsp2 (void *data, char *alias, int for_restart, int lock_held);
 
+#ifndef APACHE13
 static apr_lockmech_e
 get_apr_locking_mechanism ()
 {
@@ -167,6 +179,7 @@ get_apr_locking_mechanism ()
 		      name);
 	return APR_LOCK_DEFAULT;
 }
+#endif
 
 /* */
 static int
@@ -229,6 +242,7 @@ set_auto_application (cmd_parms *cmd, void *mconfig, const char *value)
 	return NULL;
 }
 
+
 static unsigned long
 parse_restart_time (const char *t)
 {
@@ -307,41 +321,80 @@ get_restart_mode (xsp_data *xsp, const char *value)
 	}
 }
 
+inline static uid_t
+apache_get_userid ()
+{
+#ifdef HAVE_UNIXD
+	return unixd_config.user_id;
+#else
+	return ap_user_id;
+#endif
+}
+
+inline static gid_t
+apache_get_groupid ()
+{
+#ifdef HAVE_UNIXD
+	return unixd_config.group_id;
+#else
+	return ap_group_id;
+#endif
+}
+
+inline static char *
+apache_get_username ()
+{
+#ifdef HAVE_UNIXD
+	return unixd_config.user_name;
+#else
+	return ap_user_name;
+#endif
+}
+
+#ifndef APACHE13
 static void
 ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 {
 	apr_status_t rv;
 	mode_t old_umask;
-#if defined (APR_HAS_USER) && defined (HAVE_UNIXD)
+#if defined (APR_HAS_USER)
 	apr_uid_t cur_uid;
 	apr_gid_t cur_gid;
 	int switch_back_to_root = 0;
 
-	if (unixd_config.user_id == -1 || unixd_config.group_id == -1) {
+	if (apache_get_userid () == -1 || apache_get_groupid () == -1) {
 		ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
 			      "The unix daemon module not initialized yet. Please make sure that "
 			      "your mod_mono module is loaded after the User/Group directives have "
 			      "been parsed. Not initializing the dashboard.");
 		return;
 	}
-	
+#endif
+
+#if defined (APR_HAS_USER)
 	if (apr_uid_current (&cur_uid, &cur_gid, p) == APR_SUCCESS && cur_uid == 0) {
 		DEBUG_PRINT (2, "Temporarily switching to target uid/gid");
 		switch_back_to_root = 1;
-		if (setegid (unixd_config.group_id) == -1)
+		if (setegid (apache_get_groupid ()) == -1)
 			ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 				      "setegid: unable to set effective group id to %u. %s",
-				      (unsigned)unixd_config.group_id, strerror (errno));
+				      (unsigned)apache_get_groupid (), strerror (errno));
 		
-		if (seteuid (unixd_config.user_id) == -1)
+		if (seteuid (apache_get_userid ()) == -1)
 			ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 				      "seteuid: unable to set effective user id to %u. %s",
-				      (unsigned)unixd_config.user_id, strerror (errno));
+				      (unsigned)apache_get_userid (), strerror (errno));
 	}
 #endif
 
 	if (!xsp->dashboard_mutex) {
 		DEBUG_PRINT (1, "creating dashboard mutex = %s", xsp->dashboard_lock_file);
+		
+		if (unlink (xsp->dashboard_lock_file) == -1 && errno != ENOENT) {
+			ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
+				      "Failed to remove dashboard mutex file '%s'; will attempt to continue. %s",
+				      xsp->dashboard_lock_file, strerror (errno));
+		}
 		rv = apr_global_mutex_create (&xsp->dashboard_mutex, xsp->dashboard_lock_file,
 					      get_apr_locking_mechanism (), p);
 		if (rv != APR_SUCCESS) {
@@ -364,43 +417,45 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 
 	if (!xsp->dashboard_shm) {
 		rv = apr_shm_attach (&xsp->dashboard_shm, xsp->dashboard_file, p);
-		if (rv != APR_SUCCESS) {
-			DEBUG_PRINT (1, "creating dashboard '%s'", xsp->dashboard_file);
+		if (rv == APR_SUCCESS) {
+			xsp->dashboard = apr_shm_baseaddr_get (xsp->dashboard_shm);
+		} else {
+			DEBUG_PRINT (1, "removing dashboard file '%s'", xsp->dashboard_file);
 			if (unlink (xsp->dashboard_file) == -1 && errno != ENOENT) {
-				ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
-					      "Failed to remove dashboard file '%s', further actions impossible. %s",
-					      xsp->dashboard_file, strerror (errno));
+				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
+				      "Failed to attach to existing dashboard, and removing dashboard file '%s' failed (%s). Further action impossible.",
+				      xsp->dashboard_file, strerror (errno));
 				goto restore_creds;
 			}
+	
+			DEBUG_PRINT (1, "creating dashboard '%s'", xsp->dashboard_file);
 			
 			old_umask = umask (0077);
 			rv = apr_shm_create (&xsp->dashboard_shm, sizeof (dashboard_data), xsp->dashboard_file, p);
 			umask (old_umask);
 			if (rv != APR_SUCCESS) {
 				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
-					      "Failed to create shared memory segment for backend '%s'",
-					      xsp->alias);
-				goto restore_creds;
+					      "Failed to create shared memory segment for backend '%s' at '%s'.",
+					      xsp->alias, xsp->dashboard_file);
 			} else {
 				rv = apr_shm_attach (&xsp->dashboard_shm, xsp->dashboard_file, p);
 				if (rv != APR_SUCCESS) {
 					ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
 						      "Failed to attach to the dashboard '%s'",
 						      xsp->dashboard_file);
-					return;
+					goto restore_creds;
 				}
           
 				xsp->dashboard = apr_shm_baseaddr_get (xsp->dashboard_shm);
 				xsp->dashboard->start_time = time (NULL);
 				xsp->dashboard->handled_requests = 0;
 				xsp->dashboard->restart_issued = 0;
-				goto restore_creds;
+				xsp->dashboard->active_requests = 0;
+				xsp->dashboard->waiting_requests = 0;
 			}
 		}
 	}
 
-	xsp->dashboard = apr_shm_baseaddr_get (xsp->dashboard_shm);
-
   restore_creds:
 #if defined (APR_HAS_USER) && !defined (WIN32)
 	if (switch_back_to_root) {
@@ -416,6 +471,7 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 	}
 #endif
 }
+#endif
 
 static int
 add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_default, int is_virtual)
@@ -424,8 +480,10 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	xsp_data *servers;
 	int nservers;
 	int i;
+#ifndef APACHE13
 	char num [8];
-
+#endif
+	
 	i = search_for_alias (alias, config);
 	if (i >= 0)
 		return i;
@@ -457,7 +515,11 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	server->is_virtual = is_virtual;
 	server->start_attempts = "3";
 	server->start_wait_time = "2";
-
+	server->no_flush = 1;
+	server->max_active_requests = "20";
+	server->max_waiting_requests = "20";
+	
+#ifndef APACHE13
 	apr_snprintf (num, sizeof (num), "%u", (unsigned)config->nservers + 1);
 	server->dashboard_file = apr_pstrcat (pool,
 					      DASHBOARD_FILE,
@@ -476,6 +538,7 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	server->restart_time = 0;
 
 	ensure_dashboard_initialized (config, server, pool);
+#endif
 	
 	nservers = config->nservers + 1;
 	servers = config->servers;
@@ -661,20 +724,26 @@ create_mono_server_config (apr_pool_t *p, server_rec *s)
 }
 
 static void
-request_send_response_from_memory (request_rec *r, char *byteArray, int size)
+request_send_response_from_memory (request_rec *r, char *byteArray, int size, int noFlush)
 {
+	DEBUG_PRINT (0, "sending from memory with%s flush", noFlush ? "out" : "");
+	
 #ifdef APACHE13
 	if (r->sent_bodyct == 0)
 		ap_send_http_header (r);
 #endif
 
 	ap_rwrite (byteArray, size, r);
+	if (!noFlush) {
+		DEBUG_PRINT (0, "flushing");
+		ap_rflush (r);
+	}
 }
 
 static void
 request_send_response_string (request_rec *r, char *byteArray)
 {
-	request_send_response_from_memory (r, byteArray, strlen (byteArray));
+	request_send_response_from_memory (r, byteArray, strlen (byteArray), 0);
 }
 
 /* Not connection because actual port will vary depending on Apache configuration */
@@ -950,7 +1019,7 @@ get_client_block_buffer (request_rec *r, uint32_t requested_size, uint32_t *actu
 }
 
 static int
-do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
+do_command (int command, apr_socket_t *sock, request_rec *r, int *result, xsp_data *xsp)
 {
 	int32_t size;
 	char *str;
@@ -968,7 +1037,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 		return FALSE;
 	}
 
-	DEBUG_PRINT (2, "Command received: %s", cmdNames [command]);
+	DEBUG_PRINT (2, "Command received: %s (%u)", cmdNames [command], command);
 	*result = OK;
 	switch (command) {
 		case SEND_FROM_MEMORY:
@@ -979,7 +1048,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 				apr_pool_destroy (temp_pool);
 				break;
 			}
-			request_send_response_from_memory (r, str, size);
+			request_send_response_from_memory (r, str, size, xsp->no_flush);
 			apr_pool_destroy (temp_pool);
 			break;
 		case GET_SERVER_VARIABLES:
@@ -1098,6 +1167,17 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 			if (status == -1)
 				error_message = "failed to send file (file data)";
 			break;
+
+
+		case SET_CONFIGURATION: {
+			if (read_data (sock, &xsp->no_flush, sizeof (xsp->no_flush)) == -1) {
+				error_message = "failed to set configuration (output buffering)";
+				status = -1;
+				break;
+			}
+			break;
+		}
+			
 		default:
 			error_message = "unknown command";
 			status = -1;
@@ -1415,11 +1495,11 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	int max_cpu_time = 0;
 	int status;
 	char is_master;
-#if defined (APR_HAS_USER) && defined (HAVE_UNIXD)
+#if defined (APR_HAS_USER)
 	apr_uid_t cur_uid;
 	apr_gid_t cur_gid;
 
-	if (unixd_config.user_id == -1 || unixd_config.group_id == -1) {
+	if (apache_get_userid () == -1 || apache_get_groupid () == -1) {
 		ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
 			      "The unix daemon module not initialized yet. Please make sure that "
 			      "your mod_mono module is loaded after the User/Group directives have "
@@ -1494,23 +1574,23 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	setsid ();
 	chdir ("/");
 
-#if defined (APR_HAS_USER) && defined (HAVE_UNIXD)
+#if defined (APR_HAS_USER)
 	/*
 	 * Make sure the backend runs with proper uid/gid if we're forking
 	 * from the module postconfig handler.
 	 */
 	if (apr_uid_current (&cur_uid, &cur_gid, pool) == APR_SUCCESS && cur_uid == 0) {
-		DEBUG_PRINT (2, "switching forked process group to %u", (unsigned)unixd_config.group_id);
-		if (setgid (unixd_config.group_id) == -1)
+		DEBUG_PRINT (2, "switching forked process group to %u", (unsigned)apache_get_groupid ());
+		if (setgid (apache_get_groupid ()) == -1)
 			ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 				      "setgid: unable to set group id to %u. %s",
-				      (unsigned)unixd_config.group_id, strerror (errno));
+				      (unsigned)apache_get_groupid (), strerror (errno));
 
-		DEBUG_PRINT (2, "switching forked process user to %s", unixd_config.user_name);
-		if (setuid (unixd_config.user_id) == -1)
+		DEBUG_PRINT (2, "switching forked process user to %s", apache_get_username ());
+		if (setuid (apache_get_userid ()) == -1)
 			ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 				      "setuid: unable to set user id to %u. %s",
-				      (unsigned)unixd_config.user_id, strerror (errno));
+				      (unsigned)apache_get_userid (), strerror (errno));
 	}
 #endif
   
@@ -1790,7 +1870,7 @@ send_initial_data (request_rec *r, apr_socket_t *sock, char auto_app)
 	}
 
 	ptr = str = apr_pcalloc (r->pool, size);
-	*ptr++ = 7; /* version. Keep in sync with ModMonoRequest. */
+	*ptr++ = PROTOCOL_VERSION; /* version. Keep in sync with ModMonoRequest. */
 	ptr += write_string_to_buffer (ptr, 0, r->method);
 	if (s != NULL)
 		ptr += write_string_to_buffer (ptr, 0, (s->is_virtual ? s->server_hostname : NULL));
@@ -1822,11 +1902,126 @@ send_initial_data (request_rec *r, apr_socket_t *sock, char auto_app)
 	return 0;
 }
 
+static int
+increment_active_requests (xsp_data *conf)
+{
+	apr_status_t rv;
+	
+	int max_active_requests = atoi(conf->max_active_requests);
+	int max_waiting_requests = atoi(conf->max_waiting_requests);
+
+	/* Limit the number of concurrent requests. If no
+	 * limiting is in effect (or can't be done because
+	 * there is no dashboard), return the OK status. 
+	 * Same test as in the decrement function. */
+	if (max_active_requests == 0 || !conf->dashboard_mutex || !conf->dashboard)
+		return 1;
+		
+	rv = apr_global_mutex_lock (conf->dashboard_mutex);
+	/* Drop the request on failure to lock the dashboard because
+	 * we don't want to decrement the counter later since we couldn't
+	 * increment it here. */
+	if (rv != APR_SUCCESS)
+		return 0;
+		
+	// From here on, rv holds onto whether we still have
+	// the lock acquired, just in case some error ocurrs
+	// acquiring it during the loop.
+	if (conf->dashboard->active_requests >= max_active_requests) {
+		/* We need to wait until the active requests
+		 * go below the maximum. */
+		 
+		/* However, we won't keep more than max_waiting_req requests
+		 * waiting, which means the max number of active Apache
+		 * connections associated with this mod-mono-server
+		 * is max_active_req+max_waiting_req.
+		 */
+		if (conf->dashboard->waiting_requests >= max_waiting_requests) {
+			apr_global_mutex_unlock (conf->dashboard_mutex);
+			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
+			      "Maximum number of concurrent mod_mono requests to %s reached (%d active, %d waiting). Request dropped.",
+		    	  conf->dashboard_lock_file, max_active_requests, max_waiting_requests);
+			return 0;
+		}
+
+		ap_log_error (APLOG_MARK, APLOG_INFO, STATUS_AND_SERVER,
+		      "Maximum number of concurrent mod_mono requests to %s reached (%d). Will wait and retry.",
+		      conf->dashboard_lock_file, max_active_requests);
+		      
+		conf->dashboard->waiting_requests++;
+
+		int retries = 20;
+		while (retries-- > 0) {
+			// Release the lock, wait some time, and then re-acquire.
+			apr_global_mutex_unlock (conf->dashboard_mutex);
+			apr_sleep (500000); // 0.5 seconds
+			rv = apr_global_mutex_lock (conf->dashboard_mutex);
+		 	if (rv != APR_SUCCESS) break;
+			 	
+		 	// If the number of requests is low enough, we
+		 	// can stop waiting.
+			if (conf->dashboard->active_requests < max_active_requests)
+		 		break;
+		}
+
+		// Hopefully we haven't lost the lock, but if we have we still
+		// want to at least attempt to decrement the counter.
+		conf->dashboard->waiting_requests--;
+
+	 	// If we got to the end of the loop and still too
+	 	// many requests are going, stop processing the
+	 	// request.
+		if (rv == APR_SUCCESS && conf->dashboard->active_requests >= max_active_requests) {
+			apr_global_mutex_unlock (conf->dashboard_mutex);
+			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
+				      "Maximum number (%d) of concurrent mod_mono requests to %s reached. Droping request.",
+				      max_active_requests, conf->dashboard_lock_file);
+			return 0;
+		}
+	}
+		
+	/* If we lost the lock during the loop, drop the request
+	 * because we don't want to decrement the counter later
+	 * since we couldn't increment it here. */
+	if (rv != APR_SUCCESS)
+		return 0;
+
+	conf->dashboard->active_requests++;
+	apr_global_mutex_unlock (conf->dashboard_mutex);
+	
+	return 1;
+}
+
+static void
+decrement_active_requests (xsp_data *conf)
+{
+	apr_status_t rv;
+
+	int max_active_requests = atoi(conf->max_active_requests);
+
+	/* Check if limiting is in effect. Same test as in the
+	 * increment function. */
+	if (max_active_requests == 0 || !conf->dashboard_mutex || !conf->dashboard)
+		return;
+		
+	rv = apr_global_mutex_lock (conf->dashboard_mutex);
+	// Since we incremented the counter, even if we can't
+	// get a lock, we had better attempt to decrement it.
+		
+	conf->dashboard->active_requests--;
+
+	if (rv == APR_SUCCESS)
+		apr_global_mutex_unlock (conf->dashboard_mutex);
+}
+
 static int
 mono_execute_request (request_rec *r, char auto_app)
 {
 	apr_socket_t *sock;
 	apr_status_t rv;
+#ifndef APACHE13
+	apr_status_t rv2;
+#endif
 	int command = -1;
 	int result = FALSE;
 	apr_status_t input;
@@ -1880,6 +2075,7 @@ mono_execute_request (request_rec *r, char auto_app)
 	if (start_wait_time < 2)
 		start_wait_time = 2;
 
+#ifndef APACHE13
 	ensure_dashboard_initialized (config, conf, pconf);
 	if (conf->dashboard_mutex && !conf->dashboard_mutex_initialized_in_child) {
 		/* Avoiding to call apr_global_mutex_child_init is a hack since in certain
@@ -1902,28 +2098,39 @@ mono_execute_request (request_rec *r, char auto_app)
 		} else
 			conf->dashboard_mutex_initialized_in_child = 1;
 	}
+#endif
+	
+	rv = -1; /* avoid a warning about uninitialized value */
+	if (!increment_active_requests (conf))
+		return HTTP_SERVICE_UNAVAILABLE;
 	
 	while (connect_attempts--) {
 		rv = setup_socket (&sock, conf, r->pool);
 		DEBUG_PRINT (2, "After setup_socket");
+		// Note that rv's value after the loop ends is important.
 		if (rv != APR_SUCCESS) {
-			if (rv != -1)
-				return HTTP_SERVICE_UNAVAILABLE;
-			DEBUG_PRINT (2, "No backend found, will start a new copy.");
-
-			if (conf->dashboard_mutex)
-				rv = apr_global_mutex_lock (conf->dashboard_mutex);
-			else
-				rv = APR_SUCCESS;
-			DEBUG_PRINT (1, "Acquiring the %s lock for backend start", conf->dashboard_lock_file);
-      
-			if (rv != APR_SUCCESS) {
-				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
-					      "Failed to acquire %s lock, cannot continue starting new process",
-					      conf->dashboard_lock_file);
+			if (rv != -1) {
+				decrement_active_requests (conf);
 				return HTTP_SERVICE_UNAVAILABLE;
 			}
+			DEBUG_PRINT (2, "No backend found, will start a new copy.");
+
+#ifndef APACHE13
+			if (conf->dashboard_mutex)
+				rv2 = apr_global_mutex_lock (conf->dashboard_mutex);
+			else
+				rv2 = APR_SUCCESS;
+			DEBUG_PRINT (1, "Acquiring the %s lock for backend start", conf->dashboard_lock_file);
       
+			if (rv2 != APR_SUCCESS) {
+				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv2),
+					      "Failed to acquire %s lock, cannot continue starting new process",
+					      conf->dashboard_lock_file);
+				decrement_active_requests (conf);
+				return HTTP_SERVICE_UNAVAILABLE;
+			}
+#endif
+			
 			if (socket_name != NULL && unlink (socket_name) < 0 && errno != ENOENT)
 				ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 					      "Could not remove stale socket %s. %s. Further requests will probably fail.",
@@ -1932,22 +2139,34 @@ mono_execute_request (request_rec *r, char auto_app)
 			/* give some time for warm-up */
 			DEBUG_PRINT (2, "Started new backend, sleeping %us to let it configure", (unsigned)start_wait_time);
 			apr_sleep (apr_time_from_sec (start_wait_time));
+#ifndef APACHE13
 			if (conf->dashboard_mutex) {
-				rv = apr_global_mutex_unlock (conf->dashboard_mutex);
-				if (rv != APR_SUCCESS)
-					ap_log_error (APLOG_MARK, APLOG_ALERT, STATCODE_AND_SERVER (rv),
+				rv2 = apr_global_mutex_unlock (conf->dashboard_mutex);
+				if (rv2 != APR_SUCCESS)
+					ap_log_error (APLOG_MARK, APLOG_ALERT, STATCODE_AND_SERVER (rv2),
 						      "Failed to release %s lock, the process may deadlock!",
 						      conf->dashboard_lock_file);
+
 			}
+#endif
 		} else
 			break; /* connected */
 	}
+	
+	if (rv != APR_SUCCESS) {
+		/* Failed to connect to mod-mono-server after several attempts. */
+		ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
+			      "Failed to connect to mod-mono-server after several attempts to spawn the process.");
+		decrement_active_requests (conf);
+		return HTTP_SERVICE_UNAVAILABLE;
+	}
   
 	DEBUG_PRINT (2, "Sending init data");
 	if (send_initial_data (r, sock, auto_app) != 0) {
 		ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
 			      "Failed to send initial data. %s", strerror (errno));
 		apr_socket_close (sock);
+		decrement_active_requests (conf);
 		return HTTP_SERVICE_UNAVAILABLE;
 	}
 	
@@ -1956,7 +2175,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		input = read_data (sock, (char *) &command, sizeof (int32_t));
 		if (input == sizeof (int32_t)) {
 			command = INT_FROM_LE (command);
-			result = do_command (command, sock, r, &status);
+			result = do_command (command, sock, r, &status, conf);
 		}
 	} while (input == sizeof (int32_t) && result == TRUE);
 
@@ -1966,14 +2185,17 @@ mono_execute_request (request_rec *r, char auto_app)
 			      "Command stream corrupted, last command was %d", command);
 		status = HTTP_INTERNAL_SERVER_ERROR;
 	}
+	
+	decrement_active_requests (conf);
 
+#ifndef APACHE13
 	if (conf->restart_mode > AUTORESTART_MODE_NONE) {
 		int do_restart = 0;
 		
 		DEBUG_PRINT (2, "Auto-restart enabled for '%s', checking if restart required", conf->alias);
-		
+
 		ensure_dashboard_initialized (config, conf, pconf);
-		if (!conf->dashboard_mutex)
+		if (!conf->dashboard_mutex || !conf->dashboard)
 			return status;
 
 		rv = apr_global_mutex_lock (conf->dashboard_mutex);
@@ -1985,7 +2207,7 @@ mono_execute_request (request_rec *r, char auto_app)
 				      conf->dashboard_lock_file);
 			return status;
 		}
-
+		
 		if (conf->restart_mode == AUTORESTART_MODE_REQUESTS) {
 			conf->dashboard->handled_requests++;
 			if (conf->dashboard->handled_requests > conf->restart_requests) {
@@ -2024,6 +2246,7 @@ mono_execute_request (request_rec *r, char auto_app)
 				      "Failed to release %s lock after auto-restart check, the process may deadlock!",
 				      conf->dashboard_lock_file);
 	}
+#endif
 	
 	DEBUG_PRINT (2, "Done. Status: %d", status);
 	return status;
@@ -2132,8 +2355,10 @@ start_xsp (module_cfg *config, int is_restart, char *alias)
 		if (xsp->run_xsp && !strcasecmp (xsp->run_xsp, "false"))
 			continue;
 
+#ifdef APACHE13
 		if (xsp->status != FORK_NONE)
 			continue;
+#endif
 		
 		/* If alias isn't null, skip XSPs that don't have that alias. */
 		if (alias != NULL && strcmp (xsp->alias, alias))
@@ -2166,11 +2391,13 @@ start_xsp (module_cfg *config, int is_restart, char *alias)
 			DEBUG_PRINT (0, "forking %s", xsp->alias);
 			fork_mod_mono_server (pconf, xsp);
 			xsp->status = FORK_SUCCEEDED;
+#ifndef APACHE13
 			if (xsp->dashboard) {
 				xsp->dashboard->start_time = time (NULL);
 				xsp->dashboard->handled_requests = 0;
 				xsp->dashboard->restart_issued = 0;
 			}
+#endif
 		}
 	}
 }
@@ -2186,7 +2413,9 @@ terminate_xsp2 (void *data, char *alias, int for_restart, int lock_held)
 	char *termstr = "";
 	xsp_data *xsp;
 	int i;
+#ifndef APACHE13
 	int release_lock = 0;
+#endif
 	
 	DEBUG_PRINT (0, "Terminate XSP");
 	server = (server_rec *) data;
@@ -2218,7 +2447,8 @@ terminate_xsp2 (void *data, char *alias, int for_restart, int lock_held)
 
 			remove (fn); /* Don't bother checking error */
 		}
-		
+
+#ifndef APACHE13
 		/* destroy the dashboard */     
 		if (!for_restart && xsp->dashboard_shm) {
 			DEBUG_PRINT (0, "Destroying dashboard for %s", xsp->alias);
@@ -2231,20 +2461,14 @@ terminate_xsp2 (void *data, char *alias, int for_restart, int lock_held)
 					release_lock = 1;
 			}
 			
-			if (xsp->dashboard_shm) { // it might've been released while we had been waiting
-						  // for the lock
-				rv = apr_shm_detach (xsp->dashboard_shm);
-				if (rv != APR_SUCCESS)
-					ap_log_error (APLOG_MARK, APLOG_WARNING, STATCODE_AND_SERVER (rv),
-						      "Failed to detach from the '%s' shared memory dashboard",
-						      xsp->dashboard_file);
-				
-				rv = apr_shm_destroy (xsp->dashboard_shm);
-				if (rv != APR_SUCCESS)
-					ap_log_error (APLOG_MARK, APLOG_WARNING, STATCODE_AND_SERVER (rv),
-						      "Failed to destroy the '%s' shared memory dashboard",
-						      xsp->dashboard_file);
-			}
+			// No need to detach before destroying, and in that case we must not.
+			// But should we detach instead of destroy if we weren't the creating
+			// process?
+			rv = apr_shm_destroy (xsp->dashboard_shm);
+			if (rv != APR_SUCCESS)
+				ap_log_error (APLOG_MARK, APLOG_WARNING, STATCODE_AND_SERVER (rv),
+					      "Failed to destroy the '%s' shared memory dashboard",
+					      xsp->dashboard_file);
 			
 			if (release_lock) {
 				rv = apr_global_mutex_unlock (xsp->dashboard_mutex);
@@ -2269,8 +2493,9 @@ terminate_xsp2 (void *data, char *alias, int for_restart, int lock_held)
 		}
 		
 		xsp->status = FORK_NONE;
+#endif
 	}
-
+	
 	apr_sleep (apr_time_from_sec (1));
 	return APR_SUCCESS;
 }
@@ -2290,8 +2515,10 @@ mono_control_panel_handler (request_rec *r)
 	xsp_data *xsp;
 	int i;
 	char *buffer;
+#ifndef APACHE13
 	apr_status_t rv;
-
+#endif
+	
 	if (strcmp (r->handler, "mono-ctrl"))
 		return DECLINED;
 
@@ -2319,6 +2546,7 @@ mono_control_panel_handler (request_rec *r)
 					       "Restart Server</a>", xsp->alias, xsp->alias);
 			request_send_response_string(r, buffer);
 
+#ifndef APACHE13
 			ensure_dashboard_initialized (config, xsp, pconf);
 			if (xsp->dashboard_mutex) {
 				rv = apr_global_mutex_lock (xsp->dashboard_mutex);
@@ -2340,7 +2568,8 @@ mono_control_panel_handler (request_rec *r)
 							      xsp->dashboard_lock_file);
 				}
 			}
-
+#endif
+			
 			request_send_response_string(r, "</li>\n");
 		}
 		
@@ -2429,7 +2658,7 @@ mono_init_handler (apr_pool_t *p,
 #endif
 
 #if !defined (APR_HAS_USER) || defined (WIN32)
-static void
+void
 mono_child_init (
 #ifdef APACHE2
 	apr_pool_t *p, server_rec *s
@@ -2616,6 +2845,15 @@ static const command_rec mono_cmds [] = {
 	MAKE_CMD12 (MonoAutoRestartTime, restart_time,
 		    "Time after which the backend should be auto-restarted. The time format is: "
 		    "DD[:HH[:MM[:SS]]]. Default value: 00:12:00:00"),
+
+	MAKE_CMD12 (MonoMaxActiveRequests, max_active_requests,
+		    "The maximum number of concurrent requests mod_mono will pass off to the ASP.NET backend. "
+		    "Set to zero to turn off the limit. Default value: 20"),
+	MAKE_CMD12 (MonoMaxWaitingRequests, max_waiting_requests,
+		    "The maximum number of concurrent requests mod_mono will hold while the ASP.NET backend is busy "
+		    "with the maximum number of requests specified by MonoMaxActiveRequests. "
+		    "Requests that can't be processed or held are dropped with Service Unavailable." 
+		    "Default value: 20"),
 	{ NULL }
 };
 
