@@ -52,11 +52,17 @@
 #include "http_log.h"
 #include "http_config.h"
 
-#include "apr_shm.h"
+#if defined (APACHE13)
+#include "http_conf_globals.h"
+#else
 #if !defined (WIN32) && !defined (OS2) && !defined (BEOS) && !defined (NETWARE)
 #define HAVE_UNIXD
 #include "unixd.h"
 #endif
+#endif
+
+/* KEEP IN SYNC WITH ModMonoRequest!! */
+#define PROTOCOL_VERSION 8
 
 #ifdef APACHE13
 /* Functions needed for making Apache 1.3 module as similar
@@ -64,7 +70,7 @@ as possible to Apache 2 module, reducing ifdefs in the code itself*/
 #ifdef HAVE_HTTP_PROTOCOL_H
 #include "http_protocol.h"
 #endif
-#define STATCODE_AND_SERVER (code) NULL
+#define STATCODE_AND_SERVER(__code__) NULL
 #ifndef TRUE
 #define TRUE 1
 #define FALSE 0
@@ -144,6 +150,7 @@ extern time_t ap_restart_time;
 #include <apr_version.h>
 #include <apr_strings.h>
 #include <apr_support.h>
+#include <apr_shm.h>
 
 #if APR_MAJOR_VERSION <= 0
 #define APR_SOCKET_CREATE(sock, family, type, protocol, pool) apr_socket_create (sock, family, type, pool)
@@ -214,6 +221,7 @@ enum Cmd {
 	MYNOT_FOUND, /* apache 1.3 already defines NOT_FOUND */
 	IS_CONNECTED,
 	SEND_FILE,
+	SET_CONFIGURATION,
 	LAST_COMMAND
 };
 
@@ -230,7 +238,8 @@ static char *cmdNames [] = {
 	"DECLINE_REQUEST",
 	"NOT_FOUND",
 	"IS_CONNECTED",
-	"SEND_FILE"
+	"SEND_FILE",
+	"SET_CONFIGURATION"
 };
 
 /* Module definition */
